////////////////////////////////////////////////////////////////////////////////
//  Simple LibGfl Delphi-Demo for GflSDK Version 3.11, by
//  Ingo Neumann,  E-Mail: ingo@upstart.de
//  Irwin Scollar, E-Mail: al001@mail1.rrz.uni-koeln.de
//  AmigoJack,     E-Mail: amigojack@gmx.net
//
////////////////////////////////////////////////////////////////////////////////
//  History:
//  2010-06-14: AmigoJack
//   + function MungeFPCW() to disable FPU Exceptions
//  2010-04-10: AmigoJack
//   + try/except around gflBitmapGetEXIF() and gflBitmapGetIPTC()
//  2010-03-15: AmigoJack
//   + dynamic linking/loading/initialization
//   + list all EXIF-data
//   + list all IPTC-data
//   + list ICC-profile header
//   + list extra infos
//   + (try to) list XMP-data
//   - code redundancy removed (capsulated into procedures)
//   - removed definitions "TCol24", "TLine24", "TCol32", "TLine32", "TLine1"
//   + added accelerator-chars and shortcuts to the window menu (like CTRL+O)
//     for faster keyboard-access
//   * improved "FormatInfo_Form.pas"
//  2003-02-20: Pierre-e Gougelet
//   + Support all depth
//  2001-03-12: Irwin Scollar
//   * better 1, 4 bit support
//  2001-03-09: Ingo Neumann
//   + Support for LibGfl version 1.50
//   - ABGR_to_BGRA removed, LigGfl 1.50 now supports Windows BGRA colormodel
//  2002-03-06: Irwin Scollar
//   + Support for 1,4,8 bit colormodes
//  2001-07-10: Ingo Neumann
//   + initial creation
//
////////////////////////////////////////////////////////////////////////////////


// 2010-03-09 AmigoJack:
//   Upon creating this Form (which is the first one being created) we can also dynamically
//   load LibGfl. See "LibGfl.pas" for more info on dynamic loading. This example only
//   works if the compiler switch STATIC in "LibGfl.pas" is off (i.e. commented out) and the
//   following line (compiler switch DYNAMIC_LOAD_EXAMPLE) is on (i.e. uncommented).
{$DEFINE DYNAMIC_LOAD_EXAMPLE}


unit ImageView_Form;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ExtCtrls, StdCtrls;

type
  TImageViewForm = class(TForm)
    MainMenu: TMainMenu;
    File1: TMenuItem;
    Formats1: TMenuItem;
    Image: TImage;
    Open1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    OpenDialog: TOpenDialog;
    Help1: TMenuItem;
    About1: TMenuItem;
    Memo1: TMemo;
    Button1: TButton;
    procedure Formats1Click(Sender: TObject);
    procedure ExitClick(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    NewPalette: hPalette;
  public
  end;

var
  ImageViewForm: TImageViewForm;

////////////////////////////////////////////////////////////////////////////////
implementation

uses
  LibGfl, FormatInfo_Form;
{$R *.DFM}

function IntToPixelFormat(Bpp: Integer): TPixelFormat;
begin
  case Bpp of
     1: Result := pf1bit;
     4: Result := pf4bit;
     8: Result := pf8bit;
    15: Result := pf15bit;
    16: Result := pf16bit;
    24: Result := pf24bit;
    32: Result := pf32bit;
  else
    Result := pfCustom;
  end;
end;

function ExifFlagToStr( ExifFlag: DWord ): String;
begin
  // 2010-03-15 AmigoJack
  //   I know "Flags" means a combination of several. However, these values
  //   seem to be mutually exclusive (meaning: only one does occur)
  {result:= '';

  if ExifFlag and GFL_EXIF_IFD_0               <> 0 then result:= result+ ', 0';
  if ExifFlag and GFL_EXIF_MAIN_IFD            <> 0 then result:= result+ ', Main';
  if ExifFlag and GFL_EXIF_INTEROPERABILITY_IFD<> 0 then result:= result+ ', Interoperability';
  if ExifFlag and GFL_EXIF_IFD_THUMBNAIL       <> 0 then result:= result+ ', Thumbnail';
  if ExifFlag and GFL_EXIF_GPS_IFD             <> 0 then result:= result+ ', GPS';
  if ExifFlag and GFL_EXIF_MAKERNOTE_IFD       <> 0 then result:= result+ ', Maker Note';

  if Length( result )> 0 Delete( result, 1, 2 );}

  case ExifFlag of
    GFL_EXIF_IFD_0               : result:= '0';
    GFL_EXIF_MAIN_IFD            : result:= 'Main';
    GFL_EXIF_INTEROPERABILITY_IFD: result:= 'Interoperability';
    GFL_EXIF_IFD_THUMBNAIL       : result:= 'Thumbnail';
    GFL_EXIF_GPS_IFD             : result:= 'GPS';
    GFL_EXIF_MAKERNOTE_IFD       : result:= 'Maker Note';
    else                           result:= '';
  end;
end;

//------------------------------------------------------------------------------
procedure TImageViewForm.Formats1Click(Sender: TObject);
begin
  FormatInfoForm.Show;
end;

//------------------------------------------------------------------------------
procedure TImageViewForm.ExitClick(Sender: TObject);
begin
  Close;
end;

// 2010-06-14 AmigoJack:
//   This prevents EInvalidOp to be raised by the DLL when loading RAW images
//   (like .NEF). The code basically turns off FPU exceptions; found it at:
//   http://www.pascalgamedevelopment.com/forum/index.php?topic=845.0;wap2
//
//   At http://www.ghisler.ch/board/viewtopic.php?p=180775#180775 there is
//   another interesting point:
//     Delphi turns on floating point exceptions (e.g. division by zero) and
//     lets the user handle them. Microsoft Visual C++ turns OFF floating point
//     exceptions and simply ignores them! The problem can occur when a VC++
//     library is called by a Delphi program, and a floating point exception
//     occurs in that library. It would normally be ignored when called from a
//     C++ program, but in a Delphi program it's causing a crash.
//   This is so interesting, because we deal with an EInvalidOp and not EDivByZero
//   or EZeroDivide. If at least one floating point number is incorporated in
//   the division (like "0/0.0" or "0.0/0" or "0.0/0.0") it always leads to
//   EInvalidOp - only if both dividend and divisor are integers we would get
//   any division by zero -exception.
//
//   So I assume the "error" very well is within the GFL DLLs, where the division
//   by zero is silently ignored. Hence this function.
//
//   However - I don't know if the use of this function may lead to other
//   problems. Theoretically spoken: no. But I'm far away from being sure...
function MungeFPCW( out pwOldCW: Word ): Boolean;
var
  wTemp, wSave: Word;
begin
  Result:= False;
  asm
    fstcw wSave
  end;
  if ( (wSave and $300)<> 0 )  // Not single mode
  or ( (wSave and $3F)<>$3F )  // Exceptions enabled
  or ( (wSave and $C00)<> 0 ) then begin  // Not round to nearest mode
    asm
      mov ax, wSave
      and ax, not 300h    // single mode
      or  ax, 3fh         // disable all exceptions
      and ax, not 0C00h   // round to nearest mode
      mov wTemp, ax
      fldcw   wTemp
    end;
    Result:= True;
  end;
  pwOldCW:= wSave;
end;


//------------------------------------------------------------------------------
procedure TImageViewForm.Open1Click(Sender: TObject);
var
  finfo: GFL_FILE_INFORMATION;
  lp: GFL_LOAD_PARAMS;
  gfl_bmp: PGFL_BITMAP;
  e: GFL_ERROR;
  filename: string;
  bmp: TBitmap;

  pal: PLogPalette;
  bpp: Integer;

  pExif: PGFL_EXIF_DATA;
  pExifItem: PGFL_EXIF_ENTRY;
  i2, i3: GFL_UINT32;

  pIptc: PGFL_IPTC_DATA;
  pIptcItem: PGFL_IPTC_ENTRY;

  pi8: PGFL_UINT8;
  s1: String;
  s2: WideString;

  p1, p2: PChar;
  bCrash: Boolean;

  Old80087: Word;

  // Prepare palette and assign it
  procedure _InitPalette;
  var
    _i: Integer;
  begin
    try
      GetMem( pal, SizeOf( TLogPalette )+ SizeOf( TPaletteEntry )* 256 );
      pal.palVersion:= $300;

      case bmp.PixelFormat of
        pf1bit: begin
          pal.palNumEntries:= 2;
          for _i:= 0 to 1 do begin
            pal.palPalEntry[_i].peRed  := _i* 255;
            pal.palPalEntry[_i].peGreen:= _i* 255;
            pal.palPalEntry[_i].peBlue := _i* 255;
            pal.palPalEntry[_i].peFlags:= PC_NOCOLLAPSE;
          end;
        end;
        pf4Bit: begin
          pal.palNumEntries:= 16;
          for _i:= 0 to 15 do begin
            pal.palPalEntry[_i].peRed  := gfl_bmp^.ColorMap^.Red[_i];
            pal.palPalEntry[_i].peGreen:= gfl_bmp^.ColorMap^.Green[_i];
            pal.palPalEntry[_i].peBlue := gfl_bmp^.ColorMap^.Blue[_i];
            pal.palPalEntry[_i].peFlags:= PC_NOCOLLAPSE;
          end;
        end;
        pf8Bit: begin
          pal.palNumEntries:= 256;
          if gfl_bmp^.ColorMap<> nil then begin
            for _i:= 0 to 255 do begin
              pal.palPalEntry[_i].peRed  := gfl_bmp^.ColorMap^.Red[_i];
              pal.palPalEntry[_i].peGreen:= gfl_bmp^.ColorMap^.Green[_i];
              pal.palPalEntry[_i].peBlue := gfl_bmp^.ColorMap^.Blue[_i];
              pal.palPalEntry[_i].peFlags:= PC_NOCOLLAPSE;
            end;
          end else begin
            {PCX bug in GflLib, pcx has no color palette, so make gray palette}
            for _i:= 0 to 255 do begin
              pal.palPalEntry[_i].peRed  := _i;
              pal.palPalEntry[_i].peGreen:= _i;
              pal.palPalEntry[_i].peBlue := _i;
              pal.palPalEntry[_i].peFlags:= PC_NOCOLLAPSE;
            end;
          end;
        end;
      end;

      if NewPalette<> 0 then DeleteObject( NewPalette );
      NewPalette:= CreatePalette( pal^ );
    finally
      FreeMem( pal );
    end;

    DeleteObject( bmp.ReleasePalette );
    bmp.Palette:= NewPalette;
  end;

  // Copy Pixel Data
  procedure _Move;
  var
    _y: Integer;
  begin
    // This is the exactly the macro gflGetBitmapPtr. However, Delphi doesn't know macros
    // and calling a function each loop would downgrade performance. Hence this "direct"
    // code fits more than actually using the translated macro function.
    for _y:= 0 to gfl_bmp^.Height- 1 do Move
    ( Pointer( Integer(gfl_bmp^.data)+ (_y* Integer(gfl_bmp^.BytesPerLine)) )^
    , bmp.Scanline[_y]^
    , gfl_bmp^.BytesPerLine
    );
  end;

  // Change byte order, from "http://www.delphipraxis.net/topic148180.html"
  function Swap32( X: DWord ): DWord; register;
  asm
    bswap EAX  // Converts endian
  end;

  // Reading (for now only the header of) a ICC profile. Format from
  // "http://www.color.org/icc_specs2.xalter" and "http://www.color.org/ICC1v42_2006-05.pdf"
  procedure _readICCprofile( _pStart: PByte );
  type
    _Text4= array[0..3] of Char;
  var
    _s1, _s2: String;
    _i64: Int64;
    _i, _iSize: Integer;
    _dTags, _dOffset, _dSize: DWord;
    _pCurPos, _pCurTag: PByte;
  begin
    _pCurPos:= _pStart;  // Remember start of data

    _iSize:= Swap32( PDWord( _pCurPos )^ );
    Memo1.Lines.Add( '  Size: '+ IntToStr( _iSize ) );
    Inc( _pCurPos, 4 );

    Memo1.Lines.Add( '  Preferred CMM type: '+ _Text4( PDWord( _pCurPos )^ ) );
    Inc( _pCurPos, 4 );

    _s1:= IntToStr( _pCurPos^ )+ '.';  // Major
    Inc( _pCurPos, 1 );
    _s1:= _s1+ IntToStr( (_pCurPos^ shr 4) and $F )+ '.'+ IntToStr( _pCurPos^ and $F )+ '.';  // Minor and Bugfix
    Inc( _pCurPos, 1 );
    _s1:= _s1+ IntToStr( Swap( PWord( _pCurPos )^ ) );  // Reserved
    Inc( _pCurPos, 2 );
    Memo1.Lines.Add( '  Version: '+ _s1 );

    Memo1.Lines.Add( '  Profile/device class: '+ _Text4( PDWord( _pCurPos )^ ) );
    Inc( _pCurPos, 4 );

    Memo1.Lines.Add( '  Data colour space: '+ _Text4( PDWord( _pCurPos )^ ) );
    Inc( _pCurPos, 4 );

    Memo1.Lines.Add( '  Connection space: '+ _Text4( PDWord( _pCurPos )^ ) );
    Inc( _pCurPos, 4 );

    _s1:= IntToStr( Swap( PWord( _pCurPos )^ ) )+ '-';
    Inc( _pCurPos, 2 );
    _s1:= _s1+ IntToStr( Swap( PWord( _pCurPos )^ ) )+ '-';
    Inc( _pCurPos, 2 );
    _s1:= _s1+ IntToStr( Swap( PWord( _pCurPos )^ ) )+ ' ';
    Inc( _pCurPos, 2 );
    _s1:= _s1+ IntToStr( Swap( PWord( _pCurPos )^ ) )+ ':';
    Inc( _pCurPos, 2 );
    _s1:= _s1+ IntToStr( Swap( PWord( _pCurPos )^ ) )+ ':';
    Inc( _pCurPos, 2 );
    _s1:= _s1+ IntToStr( Swap( PWord( _pCurPos )^ ) );
    Inc( _pCurPos, 2 );
    Memo1.Lines.Add( '  Date creation: '+ _s1 );

    Memo1.Lines.Add( '  Signatur: '+ _Text4( PDWord( _pCurPos )^ ) );
    Inc( _pCurPos, 4 );

    Memo1.Lines.Add( '  Primary platform: '+ _Text4( PDWord( _pCurPos )^ ) );
    Inc( _pCurPos, 4 );

    Memo1.Lines.Add( '  Flags: '+ IntToStr( Swap32( PDWord( _pCurPos )^ ) ) );
    Inc( _pCurPos, 4 );

    Memo1.Lines.Add( '  Device manufacturer: '+ _Text4( PDWord( _pCurPos )^ ) );
    Inc( _pCurPos, 4 );

    Memo1.Lines.Add( '  Device model: '+ _Text4( PDWord( _pCurPos )^ ) );
    Inc( _pCurPos, 4 );

    Int64Rec( _i64 ).Lo:= Swap32( PDWord( _pCurPos )^ );
    Inc( _pCurPos, 4 );
    Int64Rec( _i64 ).Hi:= Swap32( PDWord( _pCurPos )^ );
    Inc( _pCurPos, 4 );
    Memo1.Lines.Add( '  Device attributes: '+ IntToStr( _i64 ) );

    Memo1.Lines.Add( '  Rendering intent: '+ IntToStr( Swap32( PDWord( _pCurPos )^ ) ) );
    Inc( _pCurPos, 4 );

    Inc( _pCurPos, 12 );  // connection space illuminant

    Memo1.Lines.Add( '  Creator: '+ _Text4( PDWord( _pCurPos )^ ) );
    Inc( _pCurPos, 4 );

    SetString( _s1, PChar( _pCurPos ), 16 );
    _s2:= '';
    for _i:= 1 to 16 do _s2:= _s2+ IntToHex( Ord( _s1[_i] ), 2 );
    Memo1.Lines.Add( '  ID (MD5): '+ _s2 );
    Inc( _pCurPos, 16 );

    Inc( _pCurPos, 28 );  // Reserved


    Dec( _iSize, 128 );  // Subtract header size
    _dTags:= Swap32( PDWord( _pCurPos )^ );  // Amount of tags

    for _i:= 1 to _dTags do begin
      Inc( _pCurPos, 4 );

      _s1:= _Text4( PDWord( _pCurPos )^ );  // Signature of tag
      Inc( _pCurPos, 4 );

      _dOffset:= Swap32( PDWord( _pCurPos )^ );  // Offset of tag relative to start of data
      Inc( _pCurPos, 4 );

      if Integer( _dOffset )> _iSize- 12 then begin
        Memo1.Lines.Add( '  ! Tag data out of bounds' );
      end else begin
        _pCurTag:= _pStart;
        Inc( _pCurTag, _dOffset );
        _dSize:= Swap32( PDWord( _pCurPos )^ );

        SetString( _s2, PChar( _pCurTag ), _dSize );
        if Copy( _s2, 1, 4 )= 'text' then begin  // Currently the only one we know how to display ;-)
          Delete( _s2, 1, 8 );
          _s2:= 'text = '+ _s2;
        end;
        Memo1.Lines.Add( '  Tag '+ _s1+ ': '+ _s2 );
      end;

      Dec( _iSize, 12 );  // Subtract tag table entry size
      if _iSize<= 0 then break;  // Safety course on bad offsets
    end;
  end;

begin
  gflEnableLZW( GFL_TRUE );

  OpenDialog.Filter:= GflFilterString;
  if not OpenDialog.Execute then Exit;

  // Load GFL Bitmap
  filename:= OpenDialog.Filename;
  Memo1.Lines.Add( '' );
  Memo1.Lines.Add( '> '+ filename );

  gflGetDefaultLoadParams( lp );
  lp.ColorModel:= GFL_BGR;
  lp.LinePadding:= 4;
  lp.Flags:= lp.Flags or GFL_LOAD_READ_ALL_COMMENT or GFL_LOAD_COMMENT or GFL_LOAD_METADATA;

  // 2010-06-14 AmigoJack
  //   This turns off FPU Exceptions like divisions by zero on floating point numbers
  MungeFPCW( Old80087 );

  e:= gflLoadBitmap( PChar( filename ), @gfl_bmp, @lp, finfo );

  if e<> GFL_NO_ERROR then begin
    Memo1.Lines.Add( '! File not readable: '+ StrPas( gflGetErrorString( e ) ) );
    Exit;
  end;

  pal:= nil;
  if gfl_bmp^.Btype= GFL_BINARY then bpp:= 1 else bpp:= gfl_bmp^.BytesPerPixel* 8;

  Memo1.Lines.Add( 'BitsPerPixel: '+ IntToStr( bpp ) );
  if not (bpp in [1, 4, 8, 24, 32]) then begin
    Memo1.Lines.Add( '! Only 1, 4, 8, 24 or 32 BitsPerPixel are supported in this demo' );
  end else begin
    // Create Delphi Bitmap. If paletted, minimize memory by setting size after pixel format
    bmp:= TBitmap.Create;
    bmp.PixelFormat:= IntToPixelFormat( bpp );
    bmp.Width:= gfl_bmp^.Width;
    bmp.Height:= gfl_bmp^.Height;
    NewPalette:= 0;
    //-------------------------------------
    //Fixed. I. Scollar al001@mail1.rrz.uni-koeln.de 6.3.2002

    case bmp.PixelFormat of
      pf1bit: begin
        //pf1bit has a bug. It's palette has only zero entries
        _InitPalette;

        //set canvas to white, since positive image usually wanted
        bmp.Canvas.Brush.Color:= clWhite;
        bmp.Canvas.FillRect( Bounds( 0, 0, bmp.Width, bmp.Height ) );

        _Move;
        Image.Picture.Bitmap:= bmp;
      end;

      pf4Bit,
      pf8Bit: begin
        _InitPalette;

        _Move;
        Image.Picture.Bitmap:= bmp;
      end;

      pf24Bit, pf32Bit: begin
        _Move;
        Image.Picture.Bitmap:= bmp;
      end;
    end;

    // Free Resources
    bmp.Free;
  end;

  // List infos
  Memo1.Lines.Add( 'Dimensions: '+ IntToStr( finfo.Width )+ 'x'+ IntToStr( finfo.Height ) );
  Memo1.Lines.Add( 'Format: '+ IntToStr( finfo.FormatIndex )+ ' = '+ finfo.FormatName );

  Memo1.Lines.Add( 'Description: '+ finfo.Description );
  // The DESCRIPTION is the only value which gives us a hint about the image format version.
  // There is no other way to determinate this. Unfortunately the given version part does
  // not have a straight pattern. Values I've encountered so far:
  // - Adobe Photoshop Document
  // - Compuserve GIF (Version 87a)
  // - Compuserve GIF (Version 89a)
  // - Compuserve GIF (Version 89a - Interlaced)
  // - Gimp XCF image file
  // - JPEG TrueColor (v1.1)
  // - JPEG TrueColor (v1.2)
  // - Portable Network Graphics
  // - TIFF
  // - Windows Bitmap
  // - Windows Icon
  // As you can see, GIF has the full-written word "Version", whereas JPEG only has a "v".
  // Moreover GIF might also include the flag "Interlaced". So it is completely upon the
  // end-user to build a method to recognize the version. My suggestion:
  // - regexp for /\(v[^)]+\)$/i
  //   explained: search from end for closing bracket and then (if any) for an opening
  //   bracket followed by a "v" (case insensitive)
  // - if found: strip first, second and last char (those are the brackets) and a 'v'
  // - if first characters are 'ersion' strip them
  // - while first char in [' '] strip it
  // - ...now we either have "87a", "89a", "89a - Interlaced", "1.1" or "1.2". What to do
  //   about "Interlaced" is your decision. For me it is not part of a version but a flag
  //   to the format itself, which should be processed somewhere else. So I would go on with...
  // - if text contains " - " then terminate string at that position

  // List more infos
  Memo1.Lines.Add( 'DPI: '+ IntToStr( finfo.Xdpi )+ 'x'+ IntToStr( finfo.Ydpi ) );
  Memo1.Lines.Add( 'Image Count: '+ IntToStr( finfo.NumberOfImages ) );
  Memo1.Lines.Add( 'Compression: '+ IntToStr( finfo.Compression )+ ' = '+ finfo.CompressionDescription );

  // List comment
  if gfl_bmp^.Comment<> nil then Memo1.Lines.Add( 'Comment: '+ StrPas( gfl_bmp^.Comment ) );

  // List all EXIFs
  if gflBitmapHasEXIF( gfl_bmp )= GFL_TRUE then begin
    Memo1.Lines.Add( 'EXIF:' );

    try
      pExif:= gflBitmapGetEXIF( gfl_bmp^, GFL_EXIF_WANT_MAKERNOTES );  // get value list
      if pExif<> nil then begin  // can still be empty
        pExifItem:= pExif^.ItemsList;  // first entry
        for i2:= 1 to pExif^.NumberOfItems do begin
          Memo1.Lines.Add
          ( '  '+ IntToHex( pExifItem^.Tag, 8 )
          + ' '+ StrPas( pExifItem^.Name )
          + ': '+ StrPas( pExifItem^.Value )
          {+ ' (Flag '+ IntToStr( pExifItem^.Flag )+ ')'}
          + ' = '+ ExifFlagToStr( pExifItem^.Flag )
          );
          Inc( pExifItem );  // advance to next entry
        end;

        gflFreeEXIF( pExif^ );  // dispose value list
      end;
    except
      on e: Exception do Memo1.Lines.Add( '! EXIF-Exception ('+ e.ClassName+ '): '+ e.Message );
    end;
  end;

  // List all IPTCs
  if gflBitmapHasIPTC( gfl_bmp )= GFL_TRUE then begin
    Memo1.Lines.Add( 'IPTC:' );

    try
      pIptc:= gflBitmapGetIPTC( gfl_bmp^ );  // get value list
      if pIptc<> nil then begin  // can still be empty
        pIptcItem:= pIptc^.ItemsList;  // first entry
        for i2:= 1 to pIptc^.NumberOfItems do begin
          if pIptcItem^.Id<> 0 then Memo1.Lines.Add  // avoid 0-indexed
          ( '  '+ IntToHex( pIptcItem^.Id, 8 )
          + ' '+ StrPas( pIptcItem^.Name )
          + ': '+ StrPas( pIptcItem^.Value )
          );
          Inc( pIptcItem );  // advance to next entry
        end;

        gflFreeIPTC( pIptc^ );  // dispose value list
      end;
    except
      on e: Exception do Memo1.Lines.Add( '! IPTC-Exception ('+ e.ClassName+ '): '+ e.Message );
    end;
  end;

  // List ICC profile
  if gflBitmapHasICCProfile( gfl_bmp )= GFL_TRUE then begin
    Memo1.Lines.Add( 'ICC profile:' );

    gflBitmapGetICCProfile( gfl_bmp, pi8, i3 );  // get data
    if pi8<> nil then begin  // can still be empty
      _readICCprofile( PByte( pi8 ) );

      {SetLength( s1, i3 );
      for i2:= 1 to i3 do begin
        s1[i2]:= chr( pi8^ );
        if s1[i2]= #0 then s1[i2]:= #32;
        Inc( pi8 );
      end;

      Memo1.Lines.Add( '  Length: '+ IntToStr( i3 )+ ' = '+ s1 );}

      gflMemoryFree( pi8 );  // dispose data
    end;
  end;

  // List additional infos, like for TIFFs
  i3:= gflGetExtraInfosCount( @finfo );
  if i3> 0 then begin
    Memo1.Lines.Add( 'Extra Info:' );

    for i2:= 0 to i3- 1 do begin
      gflGetExtraInfos( @finfo, i2, p1, p2 );
      Memo1.Lines.Add( '  '+ StrPas( p1 )+ ': '+ StrPas( p2 ) );
    end;
  end;

  bCrash:= false;
  // 2010-03-13 AmigoJack:
  //   If doing it like this the LENGTH would fit according to images with XMP-data.
  //   But the data itself given back was rubbish. So I commented this example out.
  // 2010-03-18 AmigoJack:
  //   Crashes on some JPGs with the following exception:
  //     "Access violation at address 00A94BD8 in module 'libgfl311.dll'. Read of address 012A5000"
  //   And then also in gflFreeBitmap() with this:
  //     "Access violation at address 00AB2D0D in module 'libgfl311.dll'. Read of address EEFEEF06"
  //   If this happened, any further gflLoadBitmap() results in GFL_ERROR_NO_MEMORY :-(
  //   Therefore I'm looking if exception occur and if so I try to unload/unlink the whole GFL.
  //   But upon unloading the second of both DLLs (no matter in which order) an MSVC-exception
  //   raises with:
  //     "Runtime Error! R6017 - unexpected multithread lock error"
  //   This ultimately terminates the whole application :-(
  // 2010-03-19 AmigoJack:
  //   On another system (also WinXP) only one exception occurs in gflFreeBitmap():
  //     "Access violation at address 00FD4BD8 in module 'libgfl311.dll'. Read of address 01864000"
  //   Luckily this doesn't corrupt MSVC, so I can successfully unload und re-init GFL.
  try
    if gflBitmapGetXMP( gfl_bmp, pi8, i3 )= GFL_TRUE then begin
      Memo1.Lines.Add( 'XMP:' );

      SetLength( s1, i3 );
      for i2:= 1 to i3 do begin
        s1[i2]:= chr( pi8^ );
        Inc( pi8 );
      end;

      Memo1.Lines.Add( '  Length: '+ IntToStr( i3 )+ ' = '+ s1 );

      gflMemoryFree( pi8 );
    end;
  except
    on e: Exception do begin
      bCrash:= true;
      Memo1.Lines.Add( '! XMP-Exception ('+ e.ClassName+ '): '+ e.Message );
    end;
  end;

  // Free Resources
  try
    gflFreeFileInformation( finfo );
  except
    on e: Exception do begin
      bCrash:= true;
      Memo1.Lines.Add( '! FreeFileInformation-Exception ('+ e.ClassName+ '): '+ e.Message );
    end;
  end;
  try
    gflFreeBitmap( gfl_bmp^ );
  except
    on e: Exception do begin
      bCrash:= true;
      Memo1.Lines.Add( '! FreeBitmap-Exception ('+ e.ClassName+ '): '+ e.Message );
    end;
  end;

  if bCrash then begin
    try
      s2:= '';
      Memo1.Lines.Add( '! Crash detected - unloading GFL...' );
      gfl_UnloadLib( s2 );
      if Length( s2 )> 0 then Memo1.Lines.Add( s2 );
    except
      on e: Exception do Memo1.Lines.Add( '! LibraryExit-Exception ('+ e.ClassName+ '): '+ e.Message );
    end;

    FormCreate( Sender );  // reinitialize on exceptions
  end;
end;

procedure TImageViewForm.About1Click(Sender: TObject);
begin
  MessageDlg
  ( 'Simple LibGfl Delphi-Demo for GflSDK Version '+ gflGetVersion+ ', by'#13#10#13#10
  + 'Ingo Neumann'#9'E-Mail: ingo@upstart.de'#13#10
  + 'Irwin Scollar'#9'E-Mail: al001@mail1.rrz.uni-koeln.de'#13#10
  + 'AmigoJack'#9'E-Mail: amigojack@gmx.net'
  , mtInformation
  , [mbOK]
  , 0 );
end;

procedure TImageViewForm.FormDestroy(Sender: TObject);
begin
  if NewPalette<> 0 then DeleteObject( NewPalette );
end;

procedure TImageViewForm.FormCreate(Sender: TObject);
{$IFDEF DYNAMIC_LOAD_EXAMPLE}
var
  sFail: WideString;  // Holds any possible error, separated by linebreaks (Windows format)
{$ENDIF}
begin
{$IFDEF DYNAMIC_LOAD_EXAMPLE}
  sFail:= '';

  LibGfl.gfl_LoadLib  // Call to dynamically load LibGfl
  ( '..\..\..\libW\libgfl311.dll'  // Path + File of base DLL
  , '..\..\..\libW\libgfle311.dll'  // Path + File of enhancement/extra DLL
//  ( '..\..\..\lib\libgfl330.dll'  // Path + File of base DLL
//  , '..\..\..\lib\libgfle330.dll'  // Path + File of enhancement/extra DLL
  , '..\..\..\Plugins'
  , sFail  // Errors will be appended to existing text
  );

  if Length( sFail )> 0 then begin  // Errors occured
    Memo1.Lines.Add( 'The following initialization errors occured:'#13#10+ sFail );
  end else begin
    Memo1.Lines.Add( 'GFL successfully initialized' );
  end;
{$ENDIF}
end;

procedure TImageViewForm.Button1Click(Sender: TObject);
var
  iWidth, iHeight: GFL_INT32;
  iRes: GFL_ERROR;
begin
  iRes:= gflGetTextExtent( 'Text', 'Arial', 30, 0, 0, 0, 0, 0, 0, iWidth, iHeight );
  Memo1.Lines.Add( 'Normal: '+ IntToStr( iWidth )+ 'x'+ IntToStr( iHeight )+ ' Result='+ IntToStr( iRes ) );

  iRes:= gflGetTextExtent( 'Text', 'Arial', 30, 0, 1, 0, 0, 0, 0, iWidth, iHeight );
  Memo1.Lines.Add( 'Italic: '+ IntToStr( iWidth )+ 'x'+ IntToStr( iHeight )+ ' Result='+ IntToStr( iRes ) );

  iRes:= gflGetTextExtent( 'Text', 'Arial', 30, 0, 0, 1, 0, 0, 0, iWidth, iHeight );
  Memo1.Lines.Add( 'Bold: '+ IntToStr( iWidth )+ 'x'+ IntToStr( iHeight )+ ' Result='+ IntToStr( iRes ) );
end;

end.

